/* Copyright (c) allObjects - See the file https://muet.mit-license.org LICENSE for copying permission. */

// ILI9341.js - emulating module for ILI9341 (and other, esp. Pixl LCD) controllers

// 20191022

// Uses 'Modules._eCache("<moduleName>",exports) to emulate caching on 'upload'

// ToDo: to get to a more flexible / reusable Espruino emulation, this has to
//       split into 'pure', display independent and global Graphics object class
//       and ILI9341 display driver class... (like it is in Espruino. This first
//       shot at an implementation was a quick and dirty 'just get it done' thing.
//       At the same time go for:
//       - introduce with and height and color depth (and oder options of Graphics)
//       - handling color exactly as Graphics does
//       - add Graphics.buffer / .createArrayBuffer(...) comparable implementation 

(function() { var exports = {}; // create (emulating) module load context

var G = function () {
 this._eCanvas = null;         // reference to visual DOM object
 this._eRgb = [0,0,0];         // set (foreground) color (used in styles and strokes) in [r,g,b] format
 this._eClr = "#000000";       // set (foreground) color (used in styles and strokes) in #hhhhhh format
 this._eFontSize   = 20;       // vectorFont size, BUT: if <0 (-1) then bit map font is set by ._eFnt... properties
 this._eFontFamily = "Arial, Helv";
 // ._eFnt... are only set with first setFont... and are used only when _eFontSize (vectorFont) is <0 (-1)
 this._eFntBitmap;             // A column-first, MSB-first, 1bpp bitmap containing the font bitmap
 this._eFntFirstChar;          // ASCII value of first character in the font - usually 32 (space)
 this._eFntWidths;             // The width of each character in the font. Either an integer, or a string where each character represents the
 this._eFntHeight;             // The height as an integer
 this._eDcdbg = false;         // set draw char debug (do dbg on 3+ color dsp w/ non-Black/White background)

 this._eEmulLCD = false;       // 're-using' for (esp.) Pixl's LCD controller (for now)

 this._eWidth = -1;            // set by ... and returned on .getWidth()
 this._eHeight = -1;           // set by ... and returned on .getHeight()
};

G.prototype.getWidth  = function() { return this._eWidth ; }
G.prototype.getHeight = function() { return this._eHeight; }

G.prototype.clear = function() {
    var c2 = this._eCanvas.getContext("2d");
    var w = this._eCanvas.width, h = this._eCanvas.height;
    c2.fillStyle = "#000000";
    c2.strokeStyle = "#000000";
    c2.lineWidth = 1;
    c2.beginPath();
    c2.moveTo(0,0);
    c2.lineTo(w-1,0); c2.lineTo(w-1,h-1); c2.lineTo(0,h-1);
    c2.closePath();
    c2.fill();
    return this;
  };
G.prototype.setColor = function(r,g,b) {
    var s,h,c; // for r(s), g(h), b(c) ...next in alphabeth ;-)
    this._eRgb =
        [ ((s=Math.round((r||0)*255)) > 255) ? 255 : (s < 0) ? 0 : s
        , ((h=Math.round((g||0)*255)) > 255) ? 255 : (h < 0) ? 0 : h
        , ((c=Math.round((b||0)*255)) > 255) ? 255 : (c < 0) ? 0 : c
        ];
//    if (this._eEmulLCD) { this._eRgb = (s|h|c) ? [255,255,255] : [0,0,0]      ; }
    if (this._eEmulLCD) { this._eRgb = (s|h|c) ? [0,0,0]       : [255,255,255]; }
    this._eClr = "#"
        + this._eHex2(this._eRgb[0])
        + this._eHex2(this._eRgb[1])
        + this._eHex2(this._eRgb[2])
        ;
    return this;
  };
G.prototype.drawRect = function(x1,y1,x2,y2) {
    var c2 = this._eCanvas.getContext("2d");
    c2.fillStyle = this._eClr;
    c2.strokeStyle = this._eClr;
    c2.lineWidth = 1;
    c2.beginPath();
    c2.moveTo(x1,y1);
    c2.lineTo(x2,y1);
    c2.lineTo(x2,y2);
    c2.lineTo(x1,y2);
    c2.closePath();
    c2.stroke();
//    c2.strokeRect(x1,y1,x2-x1,y2-y1);
    return this;
  };
G.prototype.fillRect = function(x1,y1,x2,y2) {
    var c2 = this._eCanvas.getContext("2d");
    c2.fillStyle = this._eClr;
    c2.strokeStyle = this._eClr;
    c2.fillRect(x1,y1,x2-x1+1,y2-y1+1);
    return this;
  };
G.prototype.drawPoly = function(vs,closed) {
    var c2 = this._eCanvas.getContext("2d");
    c2.fillStyle = this._eClr;
    c2.strokeStyle = this._eClr;
    c2.lineWidth = 2;
    c2.shadow
    c2.beginPath();
    c2.moveTo(vs[0], vs[1]);
    for (var i = 3; i < vs.length; i = i + 2) { c2.lineTo(vs[i - 1], vs[i]); }
    if (closed) { c2.lineTo(vs[0], vs[1]); }
    c2.stroke();
    return this;
  };
G.prototype.fillPoly = function(vs) {
    var c2 = this._eCanvas.getContext("2d");
    c2.fillStyle = this._eClr;
    c2.strokeStyle = this._eClr;
    c2.lineWidth = 1;
    c2.beginPath();
    c2.moveTo(vs[0], vs[1]);
    for (var i = 3; i < vs.length; i = i + 2) { c2.lineTo(vs[i-1], vs[i]); }
    c2.closePath();
    c2.stroke();
    c2.fill();
    return this;
  };
G.prototype.drawLine = function(x1,y1,x2,y2) {
    var c2 = this._eCanvas.getContext("2d");
    c2.fillStyle = this._eClr;
    c2.strokeStyle = this._eClr;
    c2.lineWidth = 1;
    c2.beginPath();
    c2.moveTo(x1,y1);
    c2.lineTo(x2,y2);
    c2.stroke();
    return this;
  };
G.prototype.setFontVector = function(size) {
    this._eFontSize = size;
    return this;
  };
G.prototype.drawString = function(t,x,y) {
    // does not wrapping yet nor any boundary checks
    var c2 = this._eCanvas.getContext("2d"), i = -1, l;
    t = ("string" == typeof t) ? t : ""+t;
    if (this._eFontSize > 0) { // vectorFont
      c2.font = this._eFont();
      c2.fillStyle = this._eClr;
      c2.strokeStyle = this._eClr;
      c2.fillText(t,x,y + Math.round(this._eFontSize * 1.0));
    } else { // bit map font
      l = t.length; while (++i<l) {
        x+=this._eDrawChr(c2,x,y,t.charAt(i));
      }
    }
    return this;
  };
G.prototype._eRegisterCanvas = function(_eCanvasId, w, h) {
    _eConsole.log("ILI9341._eRegisterCanvas() id:"+_eCanvasId,"w:"+w,"h:"+h);
    this._eCanvas = document.getElementById(_eCanvasId);
    this._eCanvas.width = w;
    this._eCanvas.height = h;
    this._eWidth = w;
    this._eHeight = h;
    return this;
  };
G.prototype._eHex2 = function(n) { // n is expected to be an integer
    return ( "0123456789abcdef".charAt(Math.floor(n / 16))
           + "0123456789abcdef".charAt(           n % 16 ) );
  };
G.prototype._eFont = function() { return ("" 
      + Math.round(this._eFontSize * 1.2) + "px " + this._eFontFamily);
  };

G.prototype.setFontCustom = function(bitmap, firstChar, widths, height) {
    _eConsole.log("ILI9341.setFontCustom()");
    this._eFontSize     = -1;
    this._eFntBitmap    = bitmap;
    this._eFntFirstChar = firstChar;
    this._eFntWidths    = widths;
    this._eFntHeight    = height;
  };

G.prototype.setFontBitmap = function() { // built-in 4x6 pixel bitmapped Font

    if ( ! G.prototype._eSetFont4x6) { // font not built and set up yet, so do it now (lazy)
      // !!! do it the way as every other font does (incl. the space / blank, from 32 on) !!!
      var bitMap = "", firstChar=32, wows=4, height=6; // wows withOrWidths
      // do bitMap (column first, MSBit first, 1 bpp font map

      var sfdbg = false // set font (bitmap) debug
        , apIdx = -1 // overall pack index - which n-th pack has been invoked over all (over all characters and rows) - pre-increment-index
        , cpIdx =  0 // character pack index - which n-th pack has been invoked in current 5 characters - post-increment-index
        , ccs   = [0,0,0,0,0] // 5 3_bitmap_character_codes (<<16,<<8,<<0) for the 5 chars in source rows
        , cc         // 3_bitmap_character_codes (<<16,<<8,<<0) representing bit map of one character
        , bsh   = 20 // bit shift to left for source to or (|) with 3_bitmap_character_code for 3_bitmap_bytes - post-decrement-index
        , u;         // short for undefined

      // --------------------- form https://github.com/espruino/Espruino/blob/master/libs/graphics/bitmap_font_4x6.c

    var ___ = (0<<1) // make it one more bit - 4 bits - just like every other font does... !!!!!
      , __X = (1<<1)
      , _X_ = (2<<1)
      , _XX = (3<<1)
      , X__ = (4<<1)
      , X_X = (5<<1)
      , XX_ = (6<<1)
      , XXX = (7<<1)
      ;

    var PACK_5_TO_16 = function (A,B,C,D,E) { // nope: ((A) | (B<<3) | (C<<6) | (D<<9) | (E<<12))
        var a // argument A|B|C|D|E
            ;
        apIdx++;
        // cpIdx++; // moved to post: cpIdx overflow (height) used at end to trigger adding char(s) to bit map string
        // bsh--;   // moved to post: bit shift left for moving individual bits from row to column orientation into 3_bitmap_bytes
        if (sfdbg) { _eConsole.log("a|c packIdxs",apIdx,cpIdx," bsh:",bsh); }
        for (var adx=0; adx<5; adx++) { a=arguments[adx];
          if (a===u) { // ...because of space / blank
            ccs[adx] = u; // clear it to not be taken into account when adding to bit map (every time... \: )
          } else { // shift left by 20,15,10 the 3 bits 842 w/ pos 3,2,1 to pos in expected 3_bitmap_bytes: 23,17,11 for first pack of set of char(s)
            ccs[adx] |= ((a&8)<<bsh) | ((a&4)<<(bsh-5)) | ((a&2)<<(bsh-10)); // 4th not needed, always 0 / space between chars
        } } // done with all 5 chars (for 1 'line' (of 8) for each of them)
        cpIdx++; // cpIdx overflow (height) used at end to trigger adding char(s) to bit map string
        bsh--;   // bit shift left for moving individual bits from row to column orientation into 3_bitmap_bytes
        if (cpIdx>=height) { // all rows of current char(s) done, add them to bit map string
          for (var cdx=0; cdx<5; cdx++) { // cdx is analog adx - argument index
            if ((cc = ccs[cdx])!==u) { // add 3 chars to bit map string
              bitMap += String.fromCharCode((cc>>16)&255)  // 1st, MSByte
                      + String.fromCharCode((cc>> 8)&255)  // 2nd, 'mid' Byte
                      + String.fromCharCode( cc     &255); // 3rd, LSByte
            } ccs[cdx] = 0; } // prepare for next character w/ 3 charcodes
          cpIdx = 0; bsh = 20; // setup for 8 packs for next 5 characters (post-increment-index and post-decrement-index)
        } // done with one set of 5 chars
      }; // /PACK_5_TO_16 function

    var LCD_FONT_4X6 = [ // nope, but kind-a (from 32 on): ] IN_FLASH_MEMORY = { // from 32 up to 127

    // get the space / blank in the same way as every other font does... !!!!!
    PACK_5_TO_16( ___ ),
    PACK_5_TO_16( ___ ),
    PACK_5_TO_16( ___ ),
    PACK_5_TO_16( ___ ),
    PACK_5_TO_16( ___ ),
    PACK_5_TO_16( ___ ),

    // for bit map source data reuse line-by-line the map extracted from
    // https://github.com/espruino/Espruino/blob/master/libs/graphics/bitmap_font_4x6.c
    // beg of extract //////////////////////////////
    PACK_5_TO_16( _X_ , X_X , X_X , _X_ , X_X ),
    PACK_5_TO_16( _X_ , X_X , XXX , XXX , __X ),
    PACK_5_TO_16( _X_ , ___ , X_X , XX_ , _X_ ),
    PACK_5_TO_16( ___ , ___ , XXX , _XX , X__ ),
    PACK_5_TO_16( _X_ , ___ , X_X , XXX , X_X ),
    PACK_5_TO_16( ___ , ___ , ___ , _X_ , ___ ),

    PACK_5_TO_16( XX_ , _X_ , _X_ , _X_ , _X_ ),
    PACK_5_TO_16( XX_ , _X_ , X__ , __X , XXX ),
    PACK_5_TO_16( _XX , ___ , X__ , __X , _X_ ),
    PACK_5_TO_16( XX_ , ___ , X__ , __X , X_X ),
    PACK_5_TO_16( XXX , ___ , _X_ , _X_ , ___ ),
    PACK_5_TO_16( ___ , ___ , ___ , ___ , ___ ),

    PACK_5_TO_16( ___ , ___ , ___ , ___ , __X ),
    PACK_5_TO_16( ___ , ___ , ___ , ___ , __X ),
    PACK_5_TO_16( _X_ , ___ , ___ , ___ , _X_ ),
    PACK_5_TO_16( XXX , ___ , XXX , ___ , X__ ),
    PACK_5_TO_16( _X_ , _X_ , ___ , _X_ , X__ ),
    PACK_5_TO_16( ___ , X__ , ___ , ___ , ___ ),

    PACK_5_TO_16( _X_ , _X_ , XX_ , XX_ , _XX ),
    PACK_5_TO_16( X_X , XX_ , __X , __X , X_X ),
    PACK_5_TO_16( X_X , _X_ , _X_ , _X_ , X_X ),
    PACK_5_TO_16( X_X , _X_ , X__ , __X , XXX ),
    PACK_5_TO_16( _X_ , XXX , XXX , XX_ , __X ),
    PACK_5_TO_16( ___ , ___ , ___ , ___ , ___ ),

    PACK_5_TO_16( XXX , _XX , XXX , _X_ , _X_ ),
    PACK_5_TO_16( X__ , X__ , __X , X_X , X_X ),
    PACK_5_TO_16( XX_ , XX_ , __X , _X_ , _XX ),
    PACK_5_TO_16( __X , X_X , _X_ , X_X , __X ),
    PACK_5_TO_16( XX_ , _X_ , _X_ , _X_ , _X_ ),
    PACK_5_TO_16( ___ , ___ , ___ , ___ , ___ ),

    PACK_5_TO_16( ___ , ___ , ___ , ___ , ___ ),
    PACK_5_TO_16( ___ , ___ , ___ , ___ , ___ ),
    PACK_5_TO_16( _X_ , _X_ , _X_ , XXX , _X_ ),
    PACK_5_TO_16( ___ , ___ , X__ , ___ , __X ),
    PACK_5_TO_16( _X_ , _X_ , _X_ , XXX , _X_ ),
    PACK_5_TO_16( ___ , X__ , ___ , ___ , ___ ),

    PACK_5_TO_16( XX_ , _X_ , _X_ , XX_ , _X_ ),
    PACK_5_TO_16( __X , X_X , X_X , X_X , X_X ),
    PACK_5_TO_16( _X_ , XXX , X_X , XX_ , X__ ),
    PACK_5_TO_16( ___ , XXX , XXX , X_X , X_X ),
    PACK_5_TO_16( _X_ , X__ , X_X , XX_ , _X_ ),
    PACK_5_TO_16( ___ , _XX , ___ , ___ , ___ ),

    PACK_5_TO_16( XX_ , XXX , XXX , _XX , X_X ),
    PACK_5_TO_16( X_X , X__ , X__ , X__ , X_X ),
    PACK_5_TO_16( X_X , XX_ , XX_ , X_X , XXX ),
    PACK_5_TO_16( X_X , X__ , X__ , X_X , X_X ),
    PACK_5_TO_16( XX_ , XXX , X__ , _XX , X_X ),
    PACK_5_TO_16( ___ , ___ , ___ , ___ , ___ ),

    PACK_5_TO_16( XXX , __X , X_X , X__ , X_X ),
    PACK_5_TO_16( _X_ , __X , XX_ , X__ , XXX ),
    PACK_5_TO_16( _X_ , __X , XX_ , X__ , XXX ),
    PACK_5_TO_16( _X_ , X_X , X_X , X__ , XXX ),
    PACK_5_TO_16( XXX , _X_ , X_X , XXX , XXX ),
    PACK_5_TO_16( ___ , ___ , ___ , ___ , ___ ),

    PACK_5_TO_16( X_X , _X_ , XX_ , _X_ , XX_ ),
    PACK_5_TO_16( XXX , X_X , X_X , X_X , X_X ),
    PACK_5_TO_16( XXX , X_X , X_X , X_X , X_X ),
    PACK_5_TO_16( XXX , X_X , XX_ , XXX , XX_ ),
    PACK_5_TO_16( X_X , _X_ , X__ , _XX , X_X ),
    PACK_5_TO_16( ___ , ___ , ___ , ___ , ___ ),

    PACK_5_TO_16( _XX , XXX , X_X , X_X , X_X ),
    PACK_5_TO_16( X__ , _X_ , X_X , X_X , X_X ),
    PACK_5_TO_16( _X_ , _X_ , X_X , X_X , XXX ),
    PACK_5_TO_16( __X , _X_ , X_X , XXX , XXX ),
    PACK_5_TO_16( XX_ , _X_ , XXX , _X_ , X_X ),
    PACK_5_TO_16( ___ , ___ , ___ , ___ , ___ ),

    PACK_5_TO_16( X_X , X_X , XXX , XX_ , X__ ),
    PACK_5_TO_16( X_X , X_X , __X , X__ , X__ ),
    PACK_5_TO_16( _X_ , XXX , _X_ , X__ , _X_ ),
    PACK_5_TO_16( X_X , _X_ , X__ , X__ , __X ),
    PACK_5_TO_16( X_X , _X_ , XXX , XX_ , __X ),
    PACK_5_TO_16( ___ , ___ , ___ , ___ , ___ ),

    PACK_5_TO_16( _XX , _X_ , ___ , X__ , ___ ),
    PACK_5_TO_16( __X , X_X , ___ , _X_ , XX_ ),
    PACK_5_TO_16( __X , ___ , ___ , ___ , __X ),
    PACK_5_TO_16( __X , ___ , ___ , ___ , XXX ),
    PACK_5_TO_16( _XX , ___ , ___ , ___ , XXX ),
    PACK_5_TO_16( ___ , ___ , XXX , ___ , ___ ),

    PACK_5_TO_16( X__ , ___ , __X , ___ , __X ),
    PACK_5_TO_16( XX_ , _XX , _XX , _X_ , _X_ ),
    PACK_5_TO_16( X_X , X__ , X_X , XXX , XXX ),
    PACK_5_TO_16( X_X , X__ , X_X , X__ , _X_ ),
    PACK_5_TO_16( XX_ , _XX , _XX , _XX , _X_ ),
    PACK_5_TO_16( ___ , ___ , ___ , ___ , ___ ),

    PACK_5_TO_16( ___ , X__ , _X_ , _X_ , X__ ),
    PACK_5_TO_16( _XX , XX_ , ___ , ___ , X_X ),
    PACK_5_TO_16( X_X , X_X , XX_ , XX_ , XX_ ),
    PACK_5_TO_16( _XX , X_X , _X_ , _X_ , X_X ),
    PACK_5_TO_16( __X , X_X , XXX , _X_ , X_X ),
    PACK_5_TO_16( XX_ , ___ , ___ , X__ , ___ ),

    PACK_5_TO_16( XX_ , ___ , ___ , ___ , ___ ),
    PACK_5_TO_16( _X_ , XX_ , XX_ , _X_ , XX_ ),
    PACK_5_TO_16( _X_ , XXX , X_X , X_X , X_X ),
    PACK_5_TO_16( _X_ , XXX , X_X , X_X , X_X ),
    PACK_5_TO_16( XXX , XXX , X_X , _X_ , XX_ ),
    PACK_5_TO_16( ___ , ___ , ___ , ___ , X__ ),

    PACK_5_TO_16( ___ , ___ , ___ , _X_ , ___ ),
    PACK_5_TO_16( _XX , X_X , _XX , XXX , X_X ),
    PACK_5_TO_16( X_X , XX_ , XX_ , _X_ , X_X ),
    PACK_5_TO_16( X_X , X__ , __X , _X_ , X_X ),
    PACK_5_TO_16( _XX , X__ , XX_ , __X , _XX ),
    PACK_5_TO_16( __X , ___ , ___ , ___ , ___ ),

    PACK_5_TO_16( ___ , ___ , ___ , ___ , ___ ),
    PACK_5_TO_16( X_X , X_X , X_X , X_X , XXX ),
    PACK_5_TO_16( X_X , XXX , _X_ , X_X , _X_ ),
    PACK_5_TO_16( XXX , XXX , X_X , _XX , X__ ),
    PACK_5_TO_16( _X_ , XXX , X_X , __X , XXX ),
    PACK_5_TO_16( ___ , ___ , ___ , XX_ , ___ ),

    PACK_5_TO_16( _XX , _X_ , XX_ , _XX , XXX ),
    PACK_5_TO_16( _X_ , _X_ , _X_ , XX_ , XXX ),
    PACK_5_TO_16( XX_ , _X_ , _XX , ___ , XXX ),
    PACK_5_TO_16( _X_ , _X_ , _X_ , ___ , XXX ),
    PACK_5_TO_16( _XX , _X_ , XX_ , ___ , XXX ),
    PACK_5_TO_16( ___ , ___ , ___ , ___ , XXX )
    // end of extract //////////////////////////////
    ]; LCD_FONT_4X6 = null; // discard - no need for it, was just to use the source extract

      // do width (or widths)
      wows=4;
       // establish font in Graphics' prototype
      G.prototype._eSetFont4x6 = function() {
          this.setFontCustom(bitMap,firstChar,wows,height);
        };
    } // /font build and setup

    this._eSetFont4x6(); 

  };

G.prototype._eDrawChr = function(c2,x,y,c) {
    // returns with of character done; no canvas boundary check so far (X)]
    // (consider to move from get/update/put image data to fill rectangles to - grainy - scale fonts
    var bm = this._eFntBitmap, chrPos = c.charCodeAt(0) - this._eFntFirstChar
      , ws = this._eFntWidths, h = this._eFntHeight, cdx = -1, w, b
      , bitPos = 0, bytPos, bibPos, v, rgb = this._eRgb
      , imgData, d, dpc, dbr, dbc, i, j, k
      , xs, m, r // for debug only
    if (this._eDcdbg) { _eConsole.log("._eDrawChr(): in-args:",x,y,c);
                        _eConsole.log("font-s:",(bm) ? bm.length : "undefined", this._eFntFirstChar, (ws)
                                     ? (("number" == typeof ws) ? ws : ws.length) : "undefined", h); }
    if ("number" == typeof (w = ws)) { // fixed font
      if (chrPos < 0) return w; // ----> return w - before 1st char
      if ((bytPos = Math.floor((bitPos = chrPos * w * h) / 8)) >= bm.length) return w; // ----> return w after last char
    } else { // variable width font
      if (chrPos < 0) return ws[0]; // ----> return w - before 1st char
      if (chrPos >= ws.length) return ws[0]; // ----> return - after last char (char's width)
      while (++cdx < chrPos) bitPos+=(ws.charCodeAt(cdx) * h); // find bit position
      bytPos = Math.floor(bitPos / 8);
      w = ws.charCodeAt(chrPos); // set width
    } if (this._eDcdbg) {
      m=[]; for (j=0;j<h;j++) { r=[]; for (i=0;i<w;i++) r.push("#"); m.push(r); } xs=[]; } // dbg char graph matrix init
    bibPos = bitPos - bytPos * 8;       // bit in byt (of bit map)
    v = [128,64,32,16,8,4,2,1][bibPos]; // value of bit in byte [MSB first] (of bit map)
    b = bm.charCodeAt(bytPos);          // get bit map byte
    d = (imgData = c2.getImageData(x,y,w,h)).data; // get img w*h @ x/y to update (X)
    dpr = w*4;  // 6*4=24            // image data bytes per row (base increment per col)
    dpc = h*4;  // 8*4=32 (not used) // image data bytes per col (base increment per row)
    fdr = -4;   // -4                // set col base (pre increment)
    for (i = 0; i < w; i++) {        // for all cols (x coord):
      fdr += 4; // 0|4               //   increment col
      fdc = fdr - dpr; // -24|-20    //   set row base (pre increment)
      if (this._eDcdbg) { _eConsole.log("colBegIdx:",fdc+dpr," bit:",bitPos," byt:",bytPos," v:",v," b:",("000000000"+(b.toString(2))).substr(-8)); }
      for (j = 0; j < h; j++) {      //   for all rows rows (x coord):
        fdc += dpr; // 0:24:48:..168 //     increment row
        idx = fdc - 1; //4:28:..172  //     set pixl detail base (pre increment)
        if (b&v) {                   //     update / set the pixl
          if (this._eDcdbg) { m[j][i] = "X"; rgb = [255,255,255]; xs.push(idx+1); }              // dbg char graph matrix set pixel and bit index
          for (k = 0; k < 3; k++) {  //       do the r g b color components
            d[++idx] = rgb[k]; }     //         set r|g|b color component
          d[++idx] = 255;            //         set bit w/ no transparency
        } else {                     //     skip updating / setting the pixel
          if (this._eDcdbg) { m[j][i] = "_"; rgb = [0,0,0]; xs.push(idx+1);                      // dbg char graph matrix NOT set / skipped pixel...
                              for (k = 0; k < 3; k++) d[++idx] = rgb[k]; d[++idx]=255; idx-=4; } // ... and  bit index (w/ corrector)
          idx+=4;                    //       skip pixel (rgb + alpha bytes)
        }                            //     done w/ set and unset pixel
        bitPos++; // dbg used only   //     increment bitPos and related v (~bibPos) and evtl. bytPos and map byte
        if ( ! (v=v>>1)) {           //     next bit in byte map byte w/ shift; if byte done:
          b=bm.charCodeAt(++bytPos); //     get next bit map byte and...
          v=128; }                   //     ...resume with MSB as next
      }                              //   done w/ rows
      if (this._eDcdbg) _eConsole.log("rowIdxs:","["+xs.join(",")+"]");                          // dbg row idxs 
    }                                // done w/ colums
    c2.putImageData(imgData,x,y);    // put img w*h back @ x/y to update (X)
    if (this._eDcdbg) { for (j = h-1; j >= 0; j--) _eConsole.log(m[j].join(""));   // dbg row idxs and...
                        _eConsole.log('--> "'+c+'"',w," chrPos:",chrPos," bit:",bitPos,":"); }   // ...char graph matrix out
    return w;                        // return width fort advancing x draw position (for now only draw along +x-axis )
  };

G.prototype.setPixel = function(x,y,rgbClrAry) {
    var c2 = this._eCanvas.getContext("2d")
      , imgData = c2.createImageData(1,1)
      , data = imgData
      , rgb = this._eRgb;
    data[0] = rgb[0];
    data[1] = rgb[1];
    data[2] = rgb[2];
    data[3] = 255;
    c2.putImageData(imageData,x,y);
  };
 
G.prototype.flip = function() {
    // behavior of display and flip not implemented yet... is subject to split Graphics from display driver
  };

exports.connect = function(spi, b6, b8, b7, cb) {
    _eConsole.log("ILI9341.connect(); // incl. opt. _eRegisterCanvas();");
    var g =  new G(), args;
    if (args = Modules._eeArgs["ILI9341"]) {
      g._eRegisterCanvas(args[1],args[2],args[3]); }
    setTimeout(function(){ if (cb) cb(); }, 100);
    return g;
  };

Modules._eCache("ILI9341", exports); })(); // load emulating module
